// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: sound.cpp,v 1.27 2000/08/06 09:58:50 cisc Exp $

#include "headers.h"
#include "types.h"
#include "misc.h"
#include "schedule.h"
#include "pc88/sound.h"
#include "pc88/pc88.h"
#include "pc88/config.h"
#include "opnif.h"

//#define LOGNAME "sound"
#include "diag.h"

using namespace PC8801;

// ---------------------------------------------------------------------------
//	Ej
//
Sound::Sound()
: Device(0), sslist(0), mixingbuf(0), enabled(false), cfgflg(0), refcount(0)
{
}

Sound::~Sound()
{
	Cleanup();
}

// ---------------------------------------------------------------------------
//	Ƃ
//	rate = SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
//
bool Sound::Init(PC88* pc88, uint rate, uint32 opnclock, int bufsize)
{
	pc = pc88;
	prevtime = pc->GetCPUTick();
	enabled = false;
	mixthreshold = 16;
	
	if (!SetRate(rate, opnclock, bufsize))
		return false;
	
	pc88->AddEvent(5000, this, STATIC_CAST(TimeFunc, &Sound::UpdateCounter), 0, true);
	return true;
}

// ---------------------------------------------------------------------------
//	[gݒ
//	rate:		SOUND_55K, 44100, 22050, 11025, 0 ̂ǂꂩ
//	clock:		OPN ɗ^NbN
//	bufsize:	obt@ (TvP?)
//
bool Sound::SetRate(uint rate, uint32 clock, int bufsize)
{
	if (rate == SOUND_55K)
		rate = 44100;

	for (SSNode* n = sslist; n; n = n->next)
		n->ss->SetRate(rate);
	
	enabled = false;
	samplingrate = rate;
	
	SoundBuffer::Cleanup();
	delete[] mixingbuf;	mixingbuf = 0;

	buffersize = bufsize;
	if (bufsize > 0)
	{
		if (!SoundBuffer::Init(2, bufsize))
			return false;

		mixingbuf = new int32[2 * bufsize];
		if (!mixingbuf)
			return false;

		rate50 = rate / 50;
		tdiff = 0;
		enabled = true;
	}
	return true;
}

// ---------------------------------------------------------------------------
//	Еt
//
void Sound::Cleanup()
{
	for (SSNode* n = sslist; n; )
	{
		SSNode* next = n->next;
		delete[] n;
		n = next;
	}
	sslist = 0;

	SoundBuffer::Cleanup();
	delete[] mixingbuf; mixingbuf = 0;
}

// ---------------------------------------------------------------------------
//	
//
void Sound::Mix(Sample* dest, int nsamples, Sample* dest2, int nsamples2)
{
	int mixsamples = Min(nsamples + nsamples2, buffersize);
	if (mixsamples)
	{
		LOG1("\t\t\t\tMix %d samples\n", mixsamples);
		memset(mixingbuf, 0, mixsamples * 2 * sizeof(int32));
		{
			CriticalSection::Lock lock(cs_ss);
			for (SSNode* s = sslist; s; s = s->next)
				s->ss->Mix(mixingbuf, mixsamples);
		}
		int32* src = mixingbuf;
		
		if (dofilter)
		{
			for (; nsamples>0; nsamples--)
			{
				*dest++ = Limit(FilterL(*src++), 32767, -32768);
				*dest++ = Limit(FilterR(*src++), 32767, -32768);
			}
			for (; nsamples2>0; nsamples2--)
			{
				*dest2++ = Limit(FilterL(*src++), 32767, -32768);
				*dest2++ = Limit(FilterR(*src++), 32767, -32768);
			}
		}
		else
		{
			for (; nsamples>0; nsamples--)
			{
				*dest++ = Limit(*src++, 32767, -32768);
				*dest++ = Limit(*src++, 32767, -32768);
			}
			for (; nsamples2>0; nsamples2--)
			{
				*dest2++ = Limit(*src++, 32767, -32768);
				*dest2++ = Limit(*src++, 32767, -32768);
			}
		}
	}
}

// ---------------------------------------------------------------------------
//	ݒXV
//
void Sound::ApplyConfig(const Config* config)
{
//	SetVolume(config);
	mixthreshold = (config->flags & Config::precisemixing) ? 100 : 2000;
	order = Limit((config->lpforder) / 2, maxorder, 1);

	dofilter = false;
	if (config->flag2 & Config::lpfenable)
	{
		if (1000 <= config->lpffc && config->lpffc < samplingrate / 2)
		{
			dofilter = true;
			MakeFilter(config->lpffc);
		}
	}
}

// ---------------------------------------------------------------------------
//	ǉ
//	Sound XgɁCss Ŏw肳ꂽǉC
//	ss  SetRate ĂяoD
//
//	arg:	ss		ǉ鉹 (ISoundSource)
//	ret:	S_OK, E_FAIL, E_OUTOFMEMORY
//
bool Sound::Connect(ISoundSource* ss)
{
	CriticalSection::Lock lock(cs_ss);

	// ͊ɓo^ς݂H;
	for (SSNode** n = &sslist; *n; n=&((*n)->next))
	{
		if ((*n)->ss == ss)
			return false;
	}
	
	SSNode* nn = new SSNode;
	if (nn)
	{
		*n = nn;
		nn->next = 0;
		nn->ss = ss;
//		nn->next = sslist;
//		sslist = nn;
		ss->SetRate(samplingrate);
		return true;
	}
	return false;
}

// ---------------------------------------------------------------------------
//	Xgw肳ꂽ폜
//
//	arg:	ss		폜鉹
//	ret:	S_OK, E_HANDLE
//
bool Sound::Disconnect(ISoundSource* ss)
{
	CriticalSection::Lock lock(cs_ss);
	
	for (SSNode** r = &sslist; *r; r=&((*r)->next))
	{
		if ((*r)->ss == ss)
		{
			SSNode* d = *r;
			*r = d->next;
			delete d;
			return true;
		}
	}
	return false;
}

// ---------------------------------------------------------------------------
//	XV
//	(w肳ꂽ) Mix ĂяoC݂̎Ԃ܂ōXV	
//	̓ԂςCω钼O̒iKŌĂяo
//	x̍Č\ɂȂ()D
//
//	arg:	src		XV鉹w(̎ł͖܂)
//
bool Sound::Update(ISoundSource* /*src*/)
{
	uint32 currenttime = pc->GetCPUTick();
	
	uint32 time = currenttime - prevtime;
	if (enabled && time > mixthreshold)
	{
		prevtime = currenttime;
		// nsamples = oߎ(s) * TvO[g
		// sample = ticks * rate / clock / 100000
		// sample = ticks * (rate/50) / clock / 2000

		// MulDiv(a, b, c) = (int64) a * b / c 
		int a = MulDiv(time, rate50, pc->GetEffectiveSpeed()) + tdiff;
		int samples = a / 2000;
		tdiff = a % 2000;
		
		LOG1("Store = %5d samples\n", samples);
		Put(samples);
	}
	return true;
}

// ---------------------------------------------------------------------------
//	܂ōꂽԂ́C1Tv̒[(0-1999)߂
//
int IFCALL Sound::GetSubsampleTime(ISoundSource* /*src*/)
{
	return tdiff;
}

// ---------------------------------------------------------------------------
//	IɓJE^XV
//
void IOCALL Sound::UpdateCounter(uint)
{
	if ((pc->GetCPUTick() - prevtime) > 40000)
	{
		LOG0("Update Counter\n");
		Update(0);
	}
}

// ---------------------------------------------------------------------------
//	o^[X IIR LPF
//
#ifndef M_PI
	#define M_PI 3.14159265358979323846
#endif

#define F	4096
#define FX(f)	int(f * F)

inline int Sound::FilterL(int o)
{
	for (int j=0; j<order; j++)
	{
		int p = o + (pl[j][0] * fn[j][0] + pl[j][1] * fn[j][1]) / F;
		o = (p * fn[j][2] + pl[j][0] * fn[j][3] + pl[j][1] * fn[j][2]) / F;
		pl[j][1] = pl[j][0], pl[j][0] = p;
	}
	return o;
}

inline int Sound::FilterR(int o)
{
	for (int j=0; j<order; j++)
	{
		int p = o + (pr[j][0] * fn[j][0] + pr[j][1] * fn[j][1]) / F;
		o =  (p * fn[j][2] + pr[j][0] * fn[j][3] + pr[j][1] * fn[j][2]) / F;
		pr[j][1] = pr[j][0], pr[j][0] = p;
	}
	return o;
}

void Sound::MakeFilter(int fc)
{
	double wa = tan(M_PI * fc / samplingrate);
	double wa2 = wa*wa;

	int j;
	int n = 1;
    for (j=0; j<order; j++)
    {
		double zt = cos(n * M_PI / 4 / order);
		double ia0j = 1. / (1. + 2. * wa * zt + wa2);
		
		fn[j][0] = FX(-2. * (wa2 - 1.) * ia0j);
		fn[j][1] = FX(-(1. - 2. * wa * zt + wa2) * ia0j);
		fn[j][2] = FX(wa2 * ia0j);
		fn[j][3] = FX(2. * wa2 * ia0j);
		n += 2;
    }

	for (j=0; j<order; j++)
	{
		pl[j][0] = pl[j][1] =
		pr[j][0] = pr[j][1] = 0;
	}
}
