// ---------------------------------------------------------------------------
//	M88 - PC-8801 Emulator
//	Copyright (C) cisc 1997, 1999.
// ---------------------------------------------------------------------------
//	$Id: sound.h,v 1.25 2000/06/20 23:53:29 cisc Exp $

#ifndef PC88_SOUND_H
#define PC88_SOUND_H

#include "device.h"
#include "opna.h"
#include "psg.h"
#include "soundbuf.h"

// ---------------------------------------------------------------------------

#define SOUND_55K	55555	// Special

class PC88;
class Scheduler;

namespace PC8801
{
class Sound;
class Config;
class OPNIF;

class Sound 
: public Device, public SoundBuffer, public ISoundControl
{
public:
	enum IDFunc
	{
		updatecounter = 0,
	};

public:
	Sound();
	~Sound();
	
	bool Init(PC88* pc, uint rate, uint32 clock, int bufsize);
	void Cleanup();	
	
	void ApplyConfig(const Config* config);
	bool SetRate(uint rate, uint32 clock, int bufsize);

	void IOCALL UpdateCounter(uint);
	
	bool IFCALL Connect(ISoundSource* src);
	bool IFCALL Disconnect(ISoundSource* src);
	bool IFCALL Update(ISoundSource* src=0);
	int  IFCALL GetSubsampleTime(ISoundSource* src);
	
protected:
	void Mix(Sample* d1, int s1, Sample* d2, int s2);
	
	uint samplingrate;		// TvO[g
	uint rate50;			// samplingrate / 50
	long refcount;

private:
	struct SSNode
	{
		ISoundSource* ss;
		SSNode* next;
	};

	enum { ssrev = 1, };
	struct Status
	{
		uint8 rev;
		uint8 port40;
		uint32 prevtime;
	};

	enum
	{
		maxorder = 8,			// LPF   / 2 (pole ̐?)
	};

	void MakeFilter(int cutoff);
	int FilterL(int in);
	int FilterR(int in);

	PC88* pc;
	int32* mixingbuf;
	
	uint32 prevtime;
	uint32 cfgflg;
	int tdiff;
	uint mixthreshold;
	int buffersize;

	int order;

	int fn[maxorder][4];
	int pl[maxorder][2], pr[maxorder][2];
	
	bool enabled;
	bool dofilter;
	
	SSNode* sslist;
	CriticalSection cs_ss;
};

}


#endif // PC88_SOUND_H
